// Copyright 2018 Bartosz Jaroszewski
// SPDX-License-Identifier: GPL-2.0-or-later
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

const GnomeBluetooth = imports.gi.GnomeBluetooth;
const Signals = imports.signals;
const GLib = imports.gi.GLib;
const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Utils = Me.imports.utils;

var BluetoothController = class {
    constructor() {
        this._client = new GnomeBluetooth.Client();

        if (this._client.get_devices) {
            /* >= 42 */
            this._deviceNotifyConnected = new Set();
            this._store = this._client.get_devices();
            return;
        }

        /* <= 41 */
        this._model = this._client.get_model();
    }

    enable() {
        if (this._store) {
            /* >= 42 */
            this._client.connect('notify::default-adapter', () => {
                this._deviceNotifyConnected.clear();
                this.emit('default-adapter-changed');
            });
            this._client.connect('notify::default-adapter-powered', () => {
                this._deviceNotifyConnected.clear();
                this.emit('default-adapter-changed');
            });
            this._client.connect('device-removed', (c, path) => {
                this._deviceNotifyConnected.delete(path);
                this.emit('device-deleted');
            });
            this._client.connect('device-added', (c, device) => {
                this._connectDeviceNotify(device);
                this.emit('device-inserted', this._buildDevice(device));
            });
            return;
        }

        /* <= 41 */
        this._connectSignal(this._model, 'row-changed', (arg0, arg1, iter) => {
            if (iter) {
                let device = this._buildDevice(iter);
                if (device.isDefault) {
                    this.emit('default-adapter-changed', device);
                }
                else {
                    this.emit('device-changed', device);
                }
            }

        });
        this._connectSignal(this._model, 'row-deleted', () => {
            this.emit('device-deleted');
        });
        this._connectSignal(this._model, 'row-inserted', (arg0, arg1, iter) => {
            if (iter) {
                let device = this._buildDevice(iter);
                this.emit('device-inserted', device);
            }
        });
    }

    _connectDeviceNotify(device) {
        /* >= 42 only */
        const path = device.get_object_path();

        if (this._deviceNotifyConnected.has(path))
            return;

        device.connect('notify', (device) => {
            this.emit('device-changed', this._buildDevice(device));
        });
    }

    getDevices() {
        let devices = [];

        if (this._store) {
            /* >= 42 */
            for (let i = 0; i < this._store.get_n_items(); i++) {
                let device = this._buildDevice(this._store.get_item(i));
                devices.push(device);
            }

            return devices;
        }

        /* <= 41 */
        let adapter = this._getDefaultAdapter();
        if (!adapter)
            return [];

        let [ret, iter] = this._model.iter_children(adapter);
        while (ret) {
            let device = this._buildDevice(iter);
            devices.push(device);
            ret = this._model.iter_next(iter);
        }

        return devices;
    }

    getConnectedDevices() {
        return this.getDevices().filter((device) => {
            return device.isConnected;
        });
    }

    destroy() {
        this._disconnectSignals();
    }

    _getDefaultAdapter() {
        /* <= 41 only */
        let [ret, iter] = this._model.get_iter_first();
        while (ret) {
            let isDefault = this._model.get_value(iter, GnomeBluetooth.Column.DEFAULT);
            let isPowered = this._model.get_value(iter, GnomeBluetooth.Column.POWERED);
            if (isDefault && isPowered)
                return iter;
            ret = this._model.iter_next(iter);
        }
        return null;
    }

    _buildDevice(devOrIter) {
        if (this._store) {
            /* >= 42 */
            return new BluetoothDevice(null, devOrIter);
        }

        /* <= 41 */
        return new BluetoothDevice(this._model, devOrIter);
    }
}

Signals.addSignalMethods(BluetoothController.prototype);
Utils.addSignalsHelperMethods(BluetoothController.prototype);

var BluetoothDevice = class {
    constructor(model, devOrIter) {
        this._model = model;
        this.update(devOrIter);
    }

    update(devOrIter) {
        if (!this._model) {
            /* >= 42 */
            const dev = devOrIter;
            this.name = dev.alias || dev.name;
            this.isConnected = dev.connected;
            this.isPaired = dev.paired;
            this.mac = dev.address;
            this.isDefault = false;
            return;
        }

        /* <= 41 */
        const iter = devOrIter;
        this.name = this._model.get_value(iter, GnomeBluetooth.Column.ALIAS) || this._model.get_value(iter, GnomeBluetooth.Column.NAME);
        this.isConnected = this._model.get_value(iter, GnomeBluetooth.Column.CONNECTED);
        this.isPaired = this._model.get_value(iter, GnomeBluetooth.Column.PAIRED);
        this.mac = this._model.get_value(iter, GnomeBluetooth.Column.ADDRESS);
        this.isDefault = this._model.get_value(iter, GnomeBluetooth.Column.DEFAULT);
    }

    disconnect() {
        Utils.spawn(`bluetoothctl -- disconnect ${this.mac}`)
    }

    connect() {
        Utils.spawn(`bluetoothctl -- connect ${this.mac}`)
    }

    reconnect() {
        Utils.spawn(`bluetoothctl -- disconnect ${this.mac} && sleep 7 && bluetoothctl -- connect ${this.mac}`)
    }
}
